import pytest
from pathlib import Path

from randomnwn import *
from randomnwn.fromtext import *
from networkx import single_source_dijkstra
from numpy import hypot


@pytest.fixture
def NWN():
    out = create_NWN(seed=123)
    return out


def test_NWN_type(NWN):
    assert NWN.graph["type"] == "JDA"


def test_shortest_path(NWN):
    path_len, path = single_source_dijkstra(NWN, (33,), (138,))
    ans = [(33,), (330,), (373,), (622,), (420,), (76,), (21,), (723,), (19,), 
        (232,), (123,), (422,), (308,), (166,), (406,), (53,), (736,), (138,)]

    assert path_len == len(ans) - 1
    assert path == ans


def test_benchmark_network_JDA():
    # Get benchmark file
    current_path = Path(__file__).parent.resolve()
    benchmark = current_path.joinpath("test_networks/benchmark.txt")

    # Create JDA nanowire network
    units = {"Ron": 20.0, "rho0": 22.63676, "D0": 60.0, "l0": 1.0}
    NWN = create_NWN_from_txt(str(benchmark), units=units)

    # Calculate JDA resistance
    V = 1.0
    sol = solve_network(NWN, (0,), (1,), V)
    R = V / sol[-1]
    R *= NWN.graph["units"]["Ron"]

    # Check for the correct JDA resistance
    R_JDA = 20 + 20 + 1 / (1 / (20 + 20) + 1 / 20)  # 160/3
    assert abs(R - R_JDA) < 1e-8


def test_benchmark_network_MNR():
    # Get benchmark file
    current_path = Path(__file__).parent.resolve()
    benchmark = current_path.joinpath("test_networks/benchmark.txt")

    # Create MNR nanowire network
    units = {"Ron": 20.0, "rho0": 22.63676, "D0": 60.0, "l0": 1.0}
    NWN = create_NWN_from_txt(str(benchmark), units=units)
    convert_NWN_to_MNR(NWN)

    # Calculate MNR resistance
    V = 1.0
    sol = solve_network(NWN, (0,), (1,), V)
    R = V / sol[-1]
    R *= NWN.graph["units"]["Ron"]

    # Check for the correct MNR resistance
    const = units["rho0"] / (np.pi/4 * units["D0"]**2) * 1e3
    Rin1 = const * 1.2
    Rin2 = const * hypot(0.3, 0.3)
    Rin3 = const * hypot(1.5, 1.5)
    Rin4 = const * 1.5

    R_MNR = 20 + Rin1 + 20 + Rin2 + \
        1 / (1 / (Rin3 + 20) + 1 / (20 + Rin4 + 20))    # ~74.618
    assert abs(R - R_MNR) < 1e-8


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Various dynamic models for nanowire networks.

# References
# ----------
# [1] https://doi.org/10.1038/nature06932
# [2] https://doi.org/10.1016/j.physleta.2014.08.018
# [3] https://doi.org/10.1088/0957-4484/24/38/384004
# 
# Author: Marcus Kasdorf
# Date:   July 28, 2021

import numpy as np
import networkx as nx
from typing import Callable, List, Union, Tuple
from numbers import Number

from .calculations import solve_network


def resist_func(
    NWN: nx.Graph,
    w: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    The HP group's resistance function in nondimensionalized form.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    w : ndarray or scalar
        Nondimensionalized state variable of the memristor element(s).

    Returns
    -------
    R : ndarray or scalar
        Resistance of the memristor element(s).

    """
    Roff_Ron = NWN.graph["units"]["Roff_Ron"]
    R = w * (1 - Roff_Ron) + Roff_Ron
    return R


def _HP_model_no_decay(
    t: float, 
    w: np.ndarray,
    NWN: nx.Graph,
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]],
    voltage_func: Callable,
    edge_list: list,
    start_nodes: list,
    end_nodes: list,
    window_func: Callable,
    solver: str = "spsolve",
    kwargs: dict = None
) -> np.ndarray:
    """
    Derivative of the nondimensionalized state variables `w`.

    """
    if kwargs is None:
        kwargs = dict()

    # Solve for and set resistances
    R = resist_func(NWN, w)
    attrs = {
        edge: {"conductance": 1 / R[i]} for i, edge in enumerate(edge_list)   
    }
    nx.set_edge_attributes(NWN, attrs)

    # Find applied voltage at the current time
    applied_V = voltage_func(t)

    # Solve for voltage at each node
    *V, I = solve_network(
        NWN, source_node, drain_node, applied_V, 
        "voltage", solver, **kwargs
    )
    V = np.array(V)

    # Find voltage differences
    v0 = V[start_nodes]
    v1 = V[end_nodes]
    V_delta = np.abs(v0 - v1) * np.sign(applied_V)
        
    # Find dw/dt
    dwdt = V_delta / R * window_func(w)

    return dwdt


def _HP_model_decay(
    t: float, 
    w: np.ndarray,
    NWN: nx.Graph,
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]],
    voltage_func: Callable,
    edge_list: list,
    start_nodes: list,
    end_nodes: list,
    window_func: Callable,
    solver: str = "spsolve",
    kwargs: dict = None
) -> np.ndarray:
    """
    Derivative of the nondimensionalized state variables `w` with 
    decay value `tau`.

    """
    if kwargs is None:
        kwargs = dict()

    # Solve for and set resistances
    R = resist_func(NWN, w)
    attrs = {
        edge: {"conductance": 1 / R[i]} for i, edge in enumerate(edge_list)   
    }
    nx.set_edge_attributes(NWN, attrs)

    # Find applied voltage at the current time
    applied_V = voltage_func(t)

    # Solve for voltage at each node
    *V, I = solve_network(
        NWN, source_node, drain_node, applied_V, 
        "voltage", solver, **kwargs
    )
    V = np.array(V)

    # Find voltage differences
    v0 = V[start_nodes]
    v1 = V[end_nodes]
    V_delta = np.abs(v0 - v1) * np.sign(applied_V)

    # Get decay constant
    tau = NWN.graph["tau"]
        
    # Find dw/dt
    dwdt = (V_delta / R * window_func(w)) - (w / tau)

    return dwdt


def _HP_model_chen(
    t: float, 
    y: np.ndarray,
    NWN: nx.Graph,
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]],
    voltage_func: Callable,
    edge_list: list,
    start_nodes: list,
    end_nodes: list,
    window_func: Callable,
    solver: str = "spsolve",
    kwargs: dict = None
) -> np.ndarray:
    """
    Derivative of the nondimensionalized state variables `w`, `tau`, and
    `epsilon`.

    """
    if kwargs is None:
        kwargs = dict()

    # Unpack values
    w, tau, epsilon = np.split(y, 3)
    sigma, theta, a = NWN.graph["sigma"], NWN.graph["theta"], NWN.graph["a"]

    # Solve for and set resistances
    R = resist_func(NWN, w)
    attrs = {
        edge: {"conductance": 1 / R[i]} for i, edge in enumerate(edge_list)   
    }
    nx.set_edge_attributes(NWN, attrs)

    # Find applied voltage at the current time
    applied_V = voltage_func(t)

    # Solve for voltage at each node
    *V, I = solve_network(
        NWN, source_node, drain_node, applied_V, 
        "voltage", solver, **kwargs
    )
    V = np.array(V)

    # Find voltage differences
    v0 = V[start_nodes]
    v1 = V[end_nodes]
    V_delta = np.abs(v0 - v1) * np.sign(applied_V)
        
    # Find derivatives
    l = V_delta / R
    dw_dt = (l - ((w - epsilon) / tau)) * window_func(w)
    dtau_dt = theta * l * (a - w)
    deps_dt = sigma * l * window_func(w)
    dydt = np.hstack((dw_dt, dtau_dt, deps_dt))

    return dydt


def set_chen_params(NWN: nx.Graph, sigma: Number, theta: Number, a: Number):
    NWN.graph["sigma"] = sigma
    NWN.graph["theta"] = theta
    NWN.graph["a"] = a


from setuptools import setup

namespace = {}
with open("randomnwn/version.py", "r") as f:
    exec(f.read(), namespace)

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='randomnwn',
    version=namespace["__version__"],    
    description='Modelling and analyzing random nanowire networks in Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/marcus-k/Random-NWNs',
    author='Marcus Kasdorf',
    author_email='marcus.kasdorf@ucalgary.ca',
    license='MIT License',
    packages=['randomnwn'],
    python_requires='>=3.10'
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Functions to solve nanowire networks.
# 
# Author: Marcus Kasdorf
# Date:   July 19, 2021

import numpy as np
import scipy
import networkx as nx
from networkx.linalg import laplacian_matrix
from typing import List, Tuple, Set, Union

from .nanowires import get_edge_indices


def get_connected_nodes(NWN: nx.Graph, connected: List[Tuple]) -> Set[Tuple]:
    """
    Returns a list of nodes which are connected to any of the given nodes.

    """
    nodelist = set()
    for subset in nx.connected_components(NWN):
        if any(node in subset for node in connected):
            nodelist = nodelist.union(subset)
    return nodelist


def create_matrix(
    NWN: nx.Graph,
    value_type: str = "conductance",
    source_nodes: List[Tuple] = None,
    drain_nodes: List[Tuple] = None,
    ground_nodes: bool = False
) -> scipy.sparse.csr_matrix: 
    """
    Create the Laplacian connectivity matrix.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    value_type : {"conductance", "capacitance"}, optional
        Weight to use for the Laplacian matrix. Default: "conductance".

    source_nodes : list of tuples, optional
        Only needed if ground_nodes is True to find which nodes are to be
        grounded. Default: None.

    drain_nodes : list of tuples, optional
        If a drain node is supplied, the row and column corresponding to 
        the drain node are zeros and a one is placed at the row-column
        intersection. Default: None.

    ground_nodes : bool, optional
        If true, a small value is added to the main diagonal to avoid
        singular matrices. Default: False.

    Returns
    -------
    M : csr_matrix
        Resultant sparse Laplacian matrix.

    """
    # Error check
    TYPES = ["conductance", "capacitance"]
    if value_type not in TYPES:
        raise ValueError("Invalid matrix type.")

    # Default values
    if source_nodes is None:
        source_nodes = []
    if drain_nodes is None:
        drain_nodes = []

    # Get Laplacian matrix
    nodelist = NWN.graph["node_indices"].keys()
    nodelist_len = len(nodelist)
    M = laplacian_matrix(NWN, nodelist=nodelist, weight=value_type)

    # Ground every node with a huge resistor/tiny capacitor
    if ground_nodes:
        # Get list of node indices which are not connected to an electrode
        unconnected_indices = list(
            set(NWN.graph["node_indices"].values()).difference(
                set(NWN.graph["node_indices"][node] for node in 
                    get_connected_nodes(NWN, [*source_nodes, *drain_nodes])
                )
            )
        )

        small = np.zeros(nodelist_len)
        small[unconnected_indices] = 1e-12

        # Add small value to diagonal, grounding all non-connected nodes
        M += scipy.sparse.dia_matrix(
            (small, [0]), 
            shape = (nodelist_len, nodelist_len)
        )

    # Zero each of the drain nodes' row and column
    for drain in drain_nodes:
        # Change to lil since csr sparsity changes are expensive
        M = M.tolil()
        drain_index = NWN.graph["node_indices"][drain]
        M[drain_index] = 0
        M[:, drain_index] = 0
        M[drain_index, drain_index] = 1

    return M.tocsr()


def _solver(A, z, solver, **kwargs):
    """
    Solve sparse matrix equation.

    """
    if solver == "spsolve":
        x = scipy.sparse.linalg.spsolve(A.tocsr(), z, **kwargs)
    elif solver == "minres":
        x, exit_code = scipy.sparse.linalg.minres(A, z.toarray(), **kwargs)
    elif solver == "lgmres":
        x, exit_code = scipy.sparse.linalg.lgmres(A, z.toarray(), **kwargs)
    elif solver == "gcrotmk":
        x, exit_code = scipy.sparse.linalg.gcrotmk(A, z.toarray(), **kwargs)
    else:
        raise ValueError("Not implemented solver.")
    
    return x


def _solve_voltage(
    NWN: nx.Graph, 
    voltage: float, 
    source_nodes: List[Tuple], 
    drain_nodes: List[Tuple],
    solver: str,
    **kwargs
) -> np.ndarray:
    """
    Solve for voltages at all the nodes for a given supplied voltage.

    """
    # Find node ordering and indexes
    nodelist = NWN.graph["node_indices"].keys()
    source_indices = [NWN.graph["node_indices"][node] for node in source_nodes]

    # Ground nodes only if needed
    ground_nodes = True if solver == "spsolve" else False

    G = -create_matrix(NWN, "conductance", source_nodes, drain_nodes, ground_nodes)
    B = scipy.sparse.dok_matrix((len(nodelist), len(source_indices)))
    for i, ind in enumerate(source_indices):
        B[ind, i] = 1
    C = B.T
    D = None

    A = scipy.sparse.bmat([[G, B], [C, D]])
    z = scipy.sparse.dok_matrix((len(nodelist) + len(source_indices), 1))
    z[len(nodelist):] = voltage

    out = _solver(A, z, solver, **kwargs)
    return np.array(out, copy=False)


def _solve_current(    
    NWN: nx.Graph, 
    current: float, 
    source_nodes: List[Tuple], 
    drain_nodes: List[Tuple],
    solver: str,
    **kwargs
) -> np.ndarray:
    """
    Solve for voltages at all the nodes for a given supplied current.

    """
    # Find node ordering and indexes
    nodelist = NWN.graph["node_indices"].keys()
    source_indices = [NWN.graph["node_indices"][node] for node in source_nodes]

    ground_nodes = True if solver == "spsolve" else False

    G = create_matrix(NWN, "conductance", source_nodes, drain_nodes, ground_nodes)
    z = scipy.sparse.dok_matrix((len(nodelist), 1))
    z[source_indices] = current

    out = _solver(G, z, solver, **kwargs)
    return np.array(out, copy=False)


def solve_network(
    NWN: nx.Graph, 
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]], 
    input: float,
    type: str = "voltage",
    solver: str = "spsolve",
    **kwargs
) -> np.ndarray:
    """
    Solve for the voltages of each node in a given NWN. Each drain node will 
    be grounded. If the type is "voltage", each source node will be at the 
    specified input voltage. If the type is "current", current will be sourced
    from each source node.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    source_node : tuple, or list of tuples
        Voltage/current source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    input : float
        Supplied voltage (current) in units of v0 (i0).

    type : {"voltage", "current"}, optional
        Input type. Default: "voltage".

    solver: str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    **kwargs
        Keyword arguments passed to the solver.

    Returns
    -------
    out : ndarray
        Output array containing the voltages of each node. If the input type
        is voltage, the current is also in this array as the last element.
        
    """
    # Get lists of source and drain nodes
    if isinstance(source_node, tuple):
        source_node = [source_node]
    if isinstance(drain_node, tuple):
        drain_node = [drain_node]

    # Pass to solvers
    if type == "voltage":
        out = _solve_voltage(NWN, input, source_node, drain_node, solver, **kwargs)
    elif type == "current":
        out = _solve_current(NWN, input, source_node, drain_node, solver, **kwargs)
    else:
        raise ValueError("Invalid source type.")

    return out


def solve_drain_current(
    NWN: nx.Graph, 
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]], 
    voltage: float,
    scaled: bool = False,
    solver: str = "spsolve",
    **kwargs
) -> np.ndarray:
    """
    Solve for the current through each drain node of a NWN.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage : float
        Voltage of the source nodes.

    scaled : bool, optional
        Whether or not to scaled the output by i0. Default: False.

    solver: str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    Returns
    -------
    current_array : ndarray
        Array containing the current flow through each drain node in the order
        passed.

    """
    # Get lists of source and drain nodes
    if isinstance(source_node, tuple):
        source_node = [source_node]
    if isinstance(drain_node, tuple):
        drain_node = [drain_node]

    # Preallocate output
    current_array = np.zeros(len(drain_node))

    # Solve nodes
    out = solve_network(
        NWN, source_node, drain_node, voltage, "voltage", solver, **kwargs
    )

    # Find current through each drain node
    for i, drain in enumerate(drain_node):
        I = 0
        for node in NWN.neighbors(drain):
            V = out[NWN.graph["node_indices"][node]]
            R = 1 / NWN.edges[(node, drain)]["conductance"]
            I += V / R
        current_array[i] = I

    # Scale the output if desired
    if scaled:
        current_array *= NWN.graph["units"]["i0"]

    return current_array.squeeze()


def solve_nodal_current(
    NWN: nx.Graph, 
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]], 
    voltage: float,
    scaled: bool = False,
    solver: str = "spsolve",
    **kwargs
) -> np.ndarray:
    """
    Solve for the current through each node of a NWN. It will appear that
    no current is flowing through source (drain) nodes for positive (negative)
    voltages.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage : float
        Voltage of the source nodes.

    scaled : bool, optional
        Whether or not to scaled the output by i0. Default: False.

    solver: str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    Returns
    -------
    current_array : ndarray
        Array containing the current flow through each drain node in the order
        passed.

    """
    # Get nodal voltages
    V_out = solve_network(
        NWN, source_node, drain_node, voltage, "voltage", solver, **kwargs
    )
    
    # Preallocate output
    current_array = np.zeros(len(NWN.nodes))

    # Calculate input current for each node
    for node in NWN.nodes:
        node_ind = NWN.graph["node_indices"][node]
        for edge in NWN.edges(node):
            edge0_ind = NWN.graph["node_indices"][edge[0]]
            edge1_ind = NWN.graph["node_indices"][edge[1]]
            V_delta = V_out[edge1_ind] - V_out[edge0_ind]

            # Only add current entering a node so we can see how much
            # current passes through. Else, we just get zero due to KCL.
            if V_delta > 0:
                current_array[node_ind] += (V_delta * 
                    NWN.edges[edge]["conductance"] * np.sign(voltage))

    # Scale the output if desired
    if scaled:
        current_array *= NWN.graph["units"]["i0"]

    return current_array


def solve_edge_current(
    NWN: nx.Graph, 
    source_node: Union[Tuple, List[Tuple]], 
    drain_node: Union[Tuple, List[Tuple]], 
    voltage: float,
    scaled: bool = False,
    solver: str = "spsolve",
    **kwargs
) -> np.ndarray:
    """
    Solve for the current through each node of a NWN. It will appear that
    no current is flowing through source (drain) nodes for positive (negative)
    voltages.

    Parameters
    ----------
    NWN : Graph
        Nanowire network.

    source_node : tuple, or list of tuples
        Voltage source nodes.

    drain_node : tuple, or list of tuples
        Grounded output nodes.

    voltage : float
        Voltage of the source nodes.

    scaled : bool, optional
        Whether or not to scaled the output by i0. Default: False.

    solver: str, optional
        Name of sparse matrix solving algorithm to use. Default: "spsolve".

    Returns
    -------
    I : ndarray
        Array containing the current flow through each edge.

    """
    # Get nodal voltages
    sol = solve_network(
        NWN, source_node, drain_node, voltage, "voltage", solver, **kwargs
    )

    # Get edges and conductances
    edges, G = zip(*nx.get_edge_attributes(NWN, "conductance").items())

    # Find edges indices
    start_nodes, end_nodes = get_edge_indices(NWN, edges)

    # Find current through each edges
    v0 = sol[start_nodes]
    v1 = sol[end_nodes]
    V_delta = np.abs(v0 - v1)
    I = V_delta * G

    # Scale the output if desired
    if scaled:
        I *= NWN.graph["units"]["i0"]

    return I


def scale_sol(NWN: nx.Graph, sol: np.ndarray):
    """
    Scale the voltage and current solutions by their characteristic values.

    """
    # Get parameters
    out = np.copy(sol)
    v0 = NWN.graph["units"]["v0"]
    i0 = NWN.graph["units"]["i0"]
    node_num = len(NWN.graph["node_indices"])

    # Current sources
    if node_num == len(sol):
        out *= v0

    # Voltage sources
    else:
        out[:node_num] *= v0
        out[node_num:] *= i0

    return out

